# -*- coding: utf-8 -*- 

import pyopencl as cl
import pyopencl.array as cl_array
import numpy


ctx = cl.create_some_context()
queue = cl.CommandQueue(ctx)

def str2numpyint(s):
 sa=[]
 for i in range(len(s)):
   sa.append(   ord(s[i])   )
 return numpy.array(sa , dtype= numpy.uint8)

password = "password"
lenpassword = len(password)

passwordSalt = "12345678901234567890123" # BUG APPEARS IF LENGTH OF THIS STRING IS > 23

lenpasswordSalt = len(passwordSalt)
passwordIterations = 1
keyBytes = 16

numpy_pass_global = str2numpyint(password)
numpy_pass_global_len = numpy.uint32(lenpassword)
numpy_salt = str2numpyint(passwordSalt)
numpy_saltlen = numpy.uint32(lenpasswordSalt)
numpy_passwordIterations = numpy.uint32(passwordIterations) 
numpy_out_global_len = numpy.uint32(keyBytes)

AR_out_global =[] 
for i in range(keyBytes):
   AR_out_global.append(0)
numpy_out_global = numpy.array(AR_out_global , dtype= numpy.uint8) 

mf = cl.mem_flags
pass_global_buf = cl.Buffer(ctx, mf.READ_ONLY, numpy_pass_global.nbytes)
salt_buf        = cl.Buffer(ctx, mf.READ_ONLY, numpy_salt.nbytes)
out_global_buf  = cl.Buffer(ctx, cl.mem_flags.WRITE_ONLY, numpy_out_global.nbytes)

f = open("pbkdf2_hmac_sha1_unsplit_kernel.cl", 'r')
fstr = "".join(f.readlines())

prg = cl.Program(ctx, fstr).build()

cl.enqueue_copy(queue, pass_global_buf, numpy_pass_global, is_blocking=True)
cl.enqueue_copy(queue, salt_buf, numpy_salt, is_blocking=True)

prg.pbkdf2(queue, (1,), None, pass_global_buf, numpy_pass_global_len, salt_buf, numpy_saltlen, numpy_passwordIterations, out_global_buf,numpy_out_global_len )

cl.enqueue_copy(queue, numpy_out_global, out_global_buf).wait()

opencl_key = ''.join(x.encode('hex') for x in numpy_out_global.tostring())
print "(GPU) PBKDFv2 OpenCL output saltlen "+str(lenpasswordSalt)+" : "+str(opencl_key)

########################################################################################
# NOW the BUG:
passwordSalt = "123456789012345678901234" # ITS length is 24, now
numpy_salt = str2numpyint(passwordSalt)
lenpasswordSalt = len(passwordSalt)
numpy_saltlen = numpy.uint32(lenpasswordSalt)
salt_buf        = cl.Buffer(ctx, mf.READ_ONLY, numpy_salt.nbytes)

cl.enqueue_copy(queue, pass_global_buf, numpy_pass_global, is_blocking=True)
cl.enqueue_copy(queue, salt_buf, numpy_salt, is_blocking=True)

prg.pbkdf2(queue, (1,), None, pass_global_buf, numpy_pass_global_len, salt_buf, numpy_saltlen, numpy_passwordIterations, out_global_buf,numpy_out_global_len )

cl.enqueue_copy(queue, numpy_out_global, out_global_buf).wait()

opencl_key = ''.join(x.encode('hex') for x in numpy_out_global.tostring())
print "(GPU) PBKDFv2 OpenCL output saltlen "+str(lenpasswordSalt)+" : "+str(opencl_key)


########################################################################################
import struct
import math
import hashlib
import hmac  # RFC2104
from Crypto.Cipher import XOR

################ PBKDFv2


class PBKDFv2:

    ################ init
    def __init__(self):

        # length of pseudorandom function: 20 for SHA-1, 16 for MD5
        self.hLen = 20

    ################ makeKey
    def makeKey(self, P, S, c, dkLen, digesttype='sha1'):

        # do some sanity checks
        try:
            str(P)
            str(S)
            int(c)
            float(dkLen)
            int(c)
        except:
            print "P = %s, S = %s, c = %s, dkLen = %s:" % (P, S, c, dkLen)
            raise ValueError("ERROR! Input is not correct!")

        if dkLen > ((2 ^ 32 - 1) * self.hLen):
            maxlength = (2 ^ 32 - 1) * self.hLen
            raise ValueError("ERROR! Key is to large! Maxlength is " . str(maxlength))

        l = math.ceil(dkLen / float(self.hLen))

        T = ""
        for blockindex in range(int(l)):
            T += self.F(P, S, c, blockindex, digesttype)
        DK = T[:dkLen]

        return DK

    ################ F
    def F(self, P, S, c, i, digest):

        # The pseudorandom function, PRF, used is HMAC-SHA1 (rfc2104)
        iteration = 1

        # the first iteration; P is the key, and a concatination of
        # S and blocknumber is the message
        istr = struct.pack(">I", i + 1)
        PRFMaster = hmac.new(P, digestmod=getattr(hashlib, digest))
        PRF = PRFMaster.copy()
        PRF.update(S)
        PRF.update(istr)
        U = PRF.digest()  # the first iteration
        Fbuf = U

        while iteration < c:                  # loop through all iterations
            PRF = PRFMaster.copy()
            PRF.update(U)
            U = PRF.digest()
            Fbuf = self._xor(U, Fbuf)    # XOR this new iteration with the old one
            iteration += 1
        return Fbuf

    ################ xor
    def _xor(self, a, b):
        """Performs XOR on two strings a and b"""

        if len(a) != len(b):
            raise ValueError("ERROR: Strings are of different size! %s %s" % (len(a), len(b)))

        xor = XOR.new(a)
        return xor.encrypt(b)


####################################### MAIN ################################################
pbkdf = PBKDFv2()
# pbkdf.makeKey(password, passwordSalt, passwordIterations, keyBytes, 'sha1')
print "(CPU) PBKDFv2 Python output saltlen 23 : "+str(''.join(x.encode('hex') for x in pbkdf.makeKey("password", "12345678901234567890123", 1, 16, 'sha1')))
print "(CPU) PBKDFv2 Python output saltlen 24 : "+str(''.join(x.encode('hex') for x in pbkdf.makeKey("password", "123456789012345678901234", 1, 16, 'sha1')))
